@@ -21,6 +21,7 @@
 #include "mangle.h"
 #include "canonicalize.h"
 #include "env.h"
+#include "match.h"
 
 int endswith(const char *s, const char *sx)
 {
@@ -53,43 +54,87 @@ int startswith(const char *s, const char *sx)
         return !strncmp(s, sx, off);
 }
 
+/*
+ * Return 1 if the file does not accessible of empty
+ */
+int is_file_empty(const char *name)
+{
+	struct stat st;
+	assert(name);
+
+	return (stat(name, &st) != 0 || st.st_size == 0);
+}
+
+int mnt_parse_offset(const char *str, size_t len, uintmax_t *res)
+{
+	char *p;
+	int rc = 0;
+
+	if (!str || !*str)
+		return -EINVAL;
+
+	p = strndup(str, len);
+	if (!p)
+		return -errno;
+
+	if (strtosize(p, res))
+		rc = -EINVAL;
+	free(p);
+	return rc;
+}
+
+/* used as a callback by bsearch in mnt_fstype_is_pseudofs() */
+static int fstype_cmp(const void *v1, const void *v2)
+{
+	const char *s1 = *(const char **)v1;
+	const char *s2 = *(const char **)v2;
+
+	return strcmp(s1, s2);
+}
+
 /* returns basename and keeps dirname in the @path, if @path is "/" (root)
  * then returns empty string */
-static char *stripoff_last_component(char *path)
+char *stripoff_last_component(char *path)
 {
 	char *p = path ? strrchr(path, '/') : NULL;
 
 	if (!p)
 		return NULL;
 	*p = '\0';
-	return ++p;
+	return p + 1;
 }
 
-/* Note that the @target has to be absolute path (so at least "/")
+/*
+ * Note that the @target has to be absolute path (so at least "/").  The
+ * @filename returns allocated buffer with last path component, for example:
+ *
+ * mnt_chdir_to_parent("/mnt/test", &buf) ==> chdir("/mnt"), buf="test"
  */
 int mnt_chdir_to_parent(const char *target, char **filename)
 {
-	char *path, *last = NULL;
+	char *buf, *parent, *last = NULL;
 	char cwd[PATH_MAX];
 	int rc = -EINVAL;
 
 	if (!target || *target != '/')
 		return -EINVAL;
 
-	path = strdup(target);
-	if (!path)
+	DBG(UTILS, mnt_debug("moving to %s parent", target));
+
+	buf = strdup(target);
+	if (!buf)
 		return -ENOMEM;
 
-	if (*(path + 1) != '\0') {
-		last = stripoff_last_component(path);
+	if (*(buf + 1) != '\0') {
+		last = stripoff_last_component(buf);
 		if (!last)
 			goto err;
 	}
-	if (!*path)
-		*path = '/';	/* root */
 
-	if (chdir(path) == -1) {
-		DBG(UTILS, mnt_debug("failed to chdir to %s: %m", path));
+	parent = buf && *buf ? buf : "/";
+
+	if (chdir(parent) == -1) {
+		DBG(UTILS, mnt_debug("failed to chdir to %s: %m", parent));
 		rc = -errno;
 		goto err;
 	}
@@ -98,25 +143,65 @@ int mnt_chdir_to_parent(const char *target, char **filename)
 		rc = -errno;
 		goto err;
 	}
-	if (strcmp(cwd, path) != 0) {
-		DBG(UTILS, mnt_debug("path moved (%s -> %s)", path, cwd));
+	if (strcmp(cwd, parent) != 0) {
+		DBG(UTILS, mnt_debug(
+		    "unexpected chdir (expected=%s, cwd=%s)", parent, cwd));
 		goto err;
 	}
 
-	DBG(CXT, mnt_debug("current directory moved to %s", path));
+	DBG(CXT, mnt_debug(
+		"current directory moved to %s [last_component='%s']",
+		parent, last));
 
-	*filename = path;
+	if (filename) {
+		*filename = buf;
 
-	if (!last || !*last)
-		memcpy(*filename, ".", 2);
-	else
-		memcpy(*filename, last, strlen(last) + 1);
+		if (!last || !*last)
+			memcpy(*filename, ".", 2);
+		else
+			memcpy(*filename, last, strlen(last) + 1);
+	}
 	return 0;
 err:
-	free(path);
+	free(buf);
 	return rc;
 }
 
+/*
+ * Check if @path is on read-only filesystem independently on file permissions.
+ */
+int mnt_is_readonly(const char *path)
+{
+	if (access(path, W_OK) == 0)
+		return 0;
+	if (errno == EROFS)
+		return 1;
+	if (errno != EACCES)
+		return 0;
+
+#ifdef HAVE_FUTIMENS
+	/*
+	 * access(2) returns EACCES on read-only FS:
+	 *
+	 * - for set-uid application if one component of the path is not
+	 *   accessible for the current rUID. (Note that euidaccess(2) does not
+	 *   check for EROFS at all).
+	 *
+	 * - for read-write filesystem with read-only VFS node (aka -o remount,ro,bind)
+	 */
+	{
+		struct timespec times[2];
+
+		times[0].tv_nsec = UTIME_NOW;	/* atime */
+		times[1].tv_nsec = UTIME_OMIT;	/* mtime */
+
+		if (utimensat(AT_FDCWD, path, times, 0) == -1)
+			return errno == EROFS;
+	}
+#endif
+	return 0;
+}
+
 /**
  * mnt_mangle:
  * @str: string
@@ -151,30 +236,43 @@ char *mnt_unmangle(const char *str)
  */
 int mnt_fstype_is_pseudofs(const char *type)
 {
-	if (!type)
-		return 0;
-	if (strcmp(type, "none")  == 0 ||
-	    strcmp(type, "proc")  == 0 ||
-	    strcmp(type, "tmpfs") == 0 ||
-	    strcmp(type, "sysfs") == 0 ||
-	    strcmp(type, "autofs") == 0 ||
-	    strcmp(type, "devpts") == 0||
-	    strcmp(type, "cgroup") == 0 ||
-	    strcmp(type, "devtmpfs") == 0 ||
-	    strcmp(type, "devfs") == 0 ||
-	    strcmp(type, "dlmfs") == 0 ||
-	    strcmp(type, "cpuset") == 0 ||
-	    strcmp(type, "securityfs") == 0 ||
-	    strcmp(type, "hugetlbfs") == 0 ||
-	    strcmp(type, "rpc_pipefs") == 0 ||
-	    strcmp(type, "fusectl") == 0 ||
-	    strcmp(type, "mqueue") == 0 ||
-	    strcmp(type, "binfmt_misc") == 0 ||
-	    strcmp(type, "fuse.gvfs-fuse-daemon") == 0 ||
-	    strcmp(type, "debugfs") == 0 ||
-	    strcmp(type, "spufs") == 0)
-		return 1;
-	return 0;
+	/* This array must remain sorted when adding new fstypes */
+	static const char *pseudofs[] = {
+		"anon_inodefs",
+		"autofs",
+		"bdev",
+		"binfmt_misc",
+		"cgroup",
+		"configfs",
+		"cpuset",
+		"debugfs",
+		"devfs",
+		"devpts",
+		"devtmpfs",
+		"dlmfs",
+		"fuse.gvfs-fuse-daemon",
+		"fusectl",
+		"hugetlbfs",
+		"mqueue",
+		"nfsd",
+		"none",
+		"pipefs",
+		"proc",
+		"pstore",
+		"ramfs",
+		"rootfs",
+		"rpc_pipefs",
+		"securityfs",
+		"sockfs",
+		"spufs",
+		"sysfs",
+		"tmpfs"
+	};
+
+	assert(type);
+
+	return !(bsearch(&type, pseudofs, ARRAY_SIZE(pseudofs),
+				sizeof(char*), fstype_cmp) == NULL);
 }
 
 /**
@@ -185,8 +283,8 @@ int mnt_fstype_is_pseudofs(const char *type)
  */
 int mnt_fstype_is_netfs(const char *type)
 {
-	if (!type)
-		return 0;
+	assert(type);
+
 	if (strcmp(type, "cifs")   == 0 ||
 	    strcmp(type, "smbfs")  == 0 ||
 	    strncmp(type,"nfs", 3) == 0 ||
@@ -218,35 +316,7 @@ int mnt_fstype_is_netfs(const char *type)
  */
 int mnt_match_fstype(const char *type, const char *pattern)
 {
-	int no = 0;		/* negated types list */
-	int len;
-	const char *p;
-
-	if (!pattern && !type)
-		return 1;
-	if (!pattern)
-		return 0;
-
-	if (!strncmp(pattern, "no", 2)) {
-		no = 1;
-		pattern += 2;
-	}
-
-	/* Does type occur in types, separated by commas? */
-	len = strlen(type);
-	p = pattern;
-	while(1) {
-		if (!strncmp(p, "no", 2) && !strncmp(p+2, type, len) &&
-		    (p[len+2] == 0 || p[len+2] == ','))
-			return 0;
-		if (strncmp(p, type, len) == 0 && (p[len] == 0 || p[len] == ','))
-			return !no;
-		p = strchr(p,',');
-		if (!p)
-			break;
-		p++;
-	}
-	return no;
+	return match_fstype(type, pattern);
 }
 
 
@@ -259,7 +329,10 @@ static int check_option(const char *haystack, size_t len,
 	const char *p;
 	int no = 0;
 
-	if (needle_len >= 2 && !strncmp(needle, "no", 2)) {
+	if (needle_len >= 1 && *needle == '+') {
+		needle++;
+		needle_len--;
+	} else if (needle_len >= 2 && !strncmp(needle, "no", 2)) {
 		no = 1;
 		needle += 2;
 		needle_len -= 2;
@@ -291,10 +364,20 @@ static int check_option(const char *haystack, size_t len,
  * Unlike fs type matching, nonetdev,user and nonetdev,nouser have
  * DIFFERENT meanings; each option is matched explicitly as specified.
  *
+ * The "no" prefix interpretation could be disable by "+" prefix, for example
+ * "+noauto" matches if @optstr literally contains "noauto" string.
+ *
  * "xxx,yyy,zzz" : "nozzz"	-> False
  *
  * "xxx,yyy,zzz" : "xxx,noeee"	-> True
  *
+ * "bar,zzz"     : "nofoo"      -> True
+ *
+ * "nofoo,bar"   : "+nofoo"     -> True
+ *
+ * "bar,zzz"     : "+nofoo"     -> False
+ *
+ *
  * Returns: 1 if pattern is matching, else 0. This function also returns 0
  *          if @pattern is NULL and @optstr is non-NULL.
  */
@@ -379,42 +462,77 @@ err:
 
 static int get_filesystems(const char *filename, char ***filesystems, const char *pattern)
 {
+	int rc = 0;
 	FILE *f;
 	char line[128];
 
-	f = fopen(filename, "r");
+	f = fopen(filename, "r" UL_CLOEXECSTR);
 	if (!f)
-		return 0;
+		return 1;
+
+	DBG(UTILS, mnt_debug("reading filesystems list from: %s", filename));
 
 	while (fgets(line, sizeof(line), f)) {
 		char name[sizeof(line)];
-		int rc;
 
 		if (*line == '#' || strncmp(line, "nodev", 5) == 0)
 			continue;
 		if (sscanf(line, " %128[^\n ]\n", name) != 1)
 			continue;
+		if (strcmp(name, "*") == 0) {
+			rc = 1;
+			break;		/* end of the /etc/filesystems */
+		}
 		if (pattern && !mnt_match_fstype(name, pattern))
 			continue;
 		rc = add_filesystem(filesystems, name);
 		if (rc)
-			return rc;
+			break;
 	}
-	return 0;
+
+	fclose(f);
+	return rc;
 }
 
+/*
+ * Always check @filesystems pointer!
+ *
+ * man mount:
+ *
+ * ...mount will try to read the file /etc/filesystems, or, if that does not
+ * exist, /proc/filesystems. All of the filesystem  types  listed  there  will
+ * be tried,  except  for  those  that  are  labeled  "nodev"  (e.g.,  devpts,
+ * proc  and  nfs).  If /etc/filesystems ends in a line with a single * only,
+ * mount will read /proc/filesystems afterwards.
+ */
 int mnt_get_filesystems(char ***filesystems, const char *pattern)
 {
 	int rc;
 
 	if (!filesystems)
 		return -EINVAL;
+
 	*filesystems = NULL;
 
 	rc = get_filesystems(_PATH_FILESYSTEMS, filesystems, pattern);
-	if (rc)
+	if (rc != 1)
 		return rc;
-	return get_filesystems(_PATH_PROC_FILESYSTEMS, filesystems, pattern);
+
+	rc = get_filesystems(_PATH_PROC_FILESYSTEMS, filesystems, pattern);
+	if (rc == 1 && *filesystems)
+		rc = 0;			/* not found /proc/filesystems */
+
+	return rc;
+}
+
+static size_t get_pw_record_size(void)
+{
+#ifdef _SC_GETPW_R_SIZE_MAX
+	long sz = sysconf(_SC_GETPW_R_SIZE_MAX);
+	if (sz > 0)
+		return sz;
+#endif
+	return 16384;
 }
 
 /*
@@ -424,16 +542,9 @@ char *mnt_get_username(const uid_t uid)
 {
         struct passwd pwd;
 	struct passwd *res;
-#ifdef _SC_GETPW_R_SIZE_MAX
-	size_t sz = sysconf(_SC_GETPW_R_SIZE_MAX);
-#else
-	size_t sz = 0;
-#endif
+	size_t sz = get_pw_record_size();
 	char *buf, *username = NULL;
 
-	if (sz <= 0)
-		sz = 16384;        /* Should be more than enough */
-
 	buf = malloc(sz);
 	if (!buf)
 		return NULL;
@@ -450,13 +561,11 @@ int mnt_get_uid(const char *username, uid_t *uid)
 	int rc = -1;
         struct passwd pwd;
 	struct passwd *pw;
-	size_t sz = sysconf(_SC_GETPW_R_SIZE_MAX);
+	size_t sz = get_pw_record_size();
 	char *buf;
 
 	if (!username || !uid)
 		return -EINVAL;
-	if (sz <= 0)
-		sz = 16384;        /* Should be more than enough */
 
 	buf = malloc(sz);
 	if (!buf)
@@ -468,6 +577,7 @@ int mnt_get_uid(const char *username, uid_t *uid)
 	} else {
 		DBG(UTILS, mnt_debug(
 			"cannot convert '%s' username to UID", username));
+		rc = errno ? -errno : -EINVAL;
 	}
 
 	free(buf);
@@ -479,13 +589,11 @@ int mnt_get_gid(const char *groupname, gid_t *gid)
 	int rc = -1;
         struct group grp;
 	struct group *gr;
-	size_t sz = sysconf(_SC_GETGR_R_SIZE_MAX);
+	size_t sz = get_pw_record_size();
 	char *buf;
 
 	if (!groupname || !gid)
 		return -EINVAL;
-	if (sz <= 0)
-		sz = 16384;        /* Should be more than enough */
 
 	buf = malloc(sz);
 	if (!buf)
@@ -497,6 +605,7 @@ int mnt_get_gid(const char *groupname, gid_t *gid)
 	} else {
 		DBG(UTILS, mnt_debug(
 			"cannot convert '%s' groupname to GID", groupname));
+		rc = errno ? -errno : -EINVAL;
 	}
 
 	free(buf);
@@ -539,8 +648,8 @@ static int try_write(const char *filename)
 	if (!filename)
 		return -EINVAL;
 
-	fd = open(filename, O_RDWR|O_CREAT, S_IWUSR| \
-					    S_IRUSR|S_IRGRP|S_IROTH);
+	fd = open(filename, O_RDWR|O_CREAT|O_CLOEXEC,
+			    S_IWUSR|S_IRUSR|S_IRGRP|S_IROTH);
 	if (fd >= 0) {
 		close(fd);
 		return 0;
@@ -654,6 +763,17 @@ done:
 	return 0;
 }
 
+/**
+ * mnt_get_swaps_path:
+ *
+ * Returns: path to /proc/swaps or $LIBMOUNT_SWAPS.
+ */
+const char *mnt_get_swaps_path(void)
+{
+	const char *p = safe_getenv("LIBMOUNT_SWAPS");
+	return p ? : _PATH_PROC_SWAPS;
+}
+
 /**
  * mnt_get_fstab_path:
  *
@@ -705,9 +825,10 @@ int mnt_open_uniq_filename(const char *filename, char **name)
 {
 	int rc, fd;
 	char *n;
+	mode_t oldmode;
 
-	assert(filename);
-
+	if (!filename)
+		return -EINVAL;
 	if (name)
 		*name = NULL;
 
@@ -715,7 +836,14 @@ int mnt_open_uniq_filename(const char *filename, char **name)
 	if (rc <= 0)
 		return -errno;
 
-	fd = mkstemp(n);
+	/* This is for very old glibc and for compatibility with Posix where is
+	 * nothing about mkstemp() mode. All sane glibc use secure mode (0600).
+	 */
+	oldmode = umask(S_IRGRP|S_IWGRP|S_IXGRP|
+			S_IROTH|S_IWOTH|S_IXOTH);
+	fd = mkostemp(n, O_RDWR|O_CREAT|O_EXCL|O_CLOEXEC);
+	umask(oldmode);
+
 	if (fd >= 0 && name)
 		*name = n;
 	else
@@ -724,12 +852,24 @@ int mnt_open_uniq_filename(const char *filename, char **name)
 	return fd < 0 ? -errno : fd;
 }
 
+/**
+ * mnt_get_mountpoint:
+ * @path: pathname
+ *
+ * This function finds the mountpoint that a given path resides in. @path
+ * should be canonicalized. The returned pointer should be freed by the caller.
+ *
+ * Returns: allocated string with target of the mounted device or NULL on error
+ */
 char *mnt_get_mountpoint(const char *path)
 {
-	char *mnt = strdup(path);
+	char *mnt;
 	struct stat st;
 	dev_t dir, base;
 
+	assert(path);
+
+	mnt = strdup(path);
 	if (!mnt)
 		return NULL;
 	if (*mnt == '/' && *(mnt + 1) == '\0')
@@ -748,7 +888,8 @@ char *mnt_get_mountpoint(const char *path)
 			goto err;
 		dir = st.st_dev;
 		if (dir != base) {
-			*(p - 1) = '/';
+			if (p > mnt)
+				*(p - 1) = '/';
 			goto done;
 		}
 		base = dir;
@@ -785,6 +926,113 @@ char *mnt_get_fs_root(const char *path, const char *mnt)
 	return res;
 }
 
+/*
+ * Search for @name kernel command parametr.
+ *
+ * Returns newly allocated string with parameter argument if the @name is
+ * specified as "name=" or returns pointer to @name or returns NULL if not
+ * found.
+ *
+ * For example cmdline: "aaa bbb=BBB ccc"
+ *
+ *	@name is "aaa"	--returns--> "aaa" (pointer to @name)
+ *	@name is "bbb=" --returns--> "BBB" (allocated)
+ *	@name is "foo"  --returns--> NULL
+ */
+char *mnt_get_kernel_cmdline_option(const char *name)
+{
+	FILE *f;
+	size_t len;
+	int val = 0;
+	char *p, *res = NULL;
+	char buf[BUFSIZ];	/* see kernel include/asm-generic/setup.h: COMMAND_LINE_SIZE */
+	const char *path = _PATH_PROC_CMDLINE;
+
+	if (!name)
+		return NULL;
+
+#ifdef TEST_PROGRAM
+	path = getenv("LIBMOUNT_KERNEL_CMDLINE");
+	if (!path)
+		path = _PATH_PROC_CMDLINE;
+#endif
+	f = fopen(path, "r" UL_CLOEXECSTR);
+	if (!f)
+		return NULL;
+
+	p = fgets(buf, sizeof(buf), f);
+	fclose(f);
+
+	if (!p || !*p || *p == '\n')
+		return NULL;
+
+	len = strlen(buf);
+	*(buf + len - 1) = '\0';	/* remove last '\n' */
+
+	len = strlen(name);
+	if (len && *(name + len - 1) == '=')
+		val = 1;
+
+	for ( ; p && *p; p++) {
+		if (!(p = strstr(p, name)))
+			break;			/* not found the option */
+		if (p != buf && !isblank(*(p - 1)))
+			continue;		/* no space before the option */
+		if (!val && *(p + len) != '\0' && !isblank(*(p + len)))
+			continue;		/* no space behind the option */
+		if (val) {
+			char *v = p + len;
+
+			while (*p && !isblank(*p))	/* jump to the end of the argument */
+				p++;
+			*p = '\0';
+			res = strdup(v);
+			break;
+		} else
+			res = (char *) name;	/* option without '=' */
+		break;
+	}
+
+	return res;
+}
+
+int mkdir_p(const char *path, mode_t mode)
+{
+	char *p, *dir;
+	int rc = 0;
+
+	if (!path || !*path)
+		return -EINVAL;
+
+	dir = p = strdup(path);
+	if (!dir)
+		return -ENOMEM;
+
+	if (*p == '/')
+		p++;
+
+	while (p && *p) {
+		char *e = strchr(p, '/');
+		if (e)
+			*e = '\0';
+		if (*p) {
+			rc = mkdir(dir, mode);
+			if (rc && errno != EEXIST)
+				break;
+			rc = 0;
+		}
+		if (!e)
+			break;
+		*e = '/';
+		p = e + 1;
+	}
+
+	DBG(UTILS, mnt_debug("%s mkdir %s", path, rc ? "FAILED" : "SUCCESS"));
+
+	free(dir);
+	return rc;
+}
+
 #ifdef TEST_PROGRAM
 int test_match_fstype(struct libmnt_test *ts, int argc, char *argv[])
 {
@@ -878,6 +1126,36 @@ int test_chdir(struct libmnt_test *ts, int argc, char *argv[])
 	return rc;
 }
 
+int test_kernel_cmdline(struct libmnt_test *ts, int argc, char *argv[])
+{
+	char *name = argv[1];
+	char *res;
+
+	res = mnt_get_kernel_cmdline_option(name);
+	if (!res)
+		printf("'%s' not found\n", name);
+	else if (res == name)
+		printf("'%s' found\n", name);
+	else {
+		printf("'%s' found, argument: '%s'\n", name, res);
+		free(res);
+	}
+
+	return 0;
+}
+
+int test_mkdir(struct libmnt_test *ts, int argc, char *argv[])
+{
+	int rc;
+
+	rc = mkdir_p(argv[1], S_IRWXU |
+			 S_IRGRP | S_IXGRP |
+			 S_IROTH | S_IXOTH);
+	if (rc)
+		printf("mkdir %s failed\n", argv[1]);
+	return rc;
+}
+
 
 int main(int argc, char *argv[])
 {
@@ -890,6 +1168,9 @@ int main(int argc, char *argv[])
 	{ "--mountpoint",    test_mountpoint,      "<path>" },
 	{ "--fs-root",       test_fsroot,          "<path>" },
 	{ "--cd-parent",     test_chdir,           "<path>" },
+	{ "--kernel-cmdline",test_kernel_cmdline,  "<option> | <option>=" },
+	{ "--mkdir",         test_mkdir,           "<path>" },
+
 	{ NULL }
 	};
 
